<!DOCTYPE html>
<html>
    <head>
        <title>Report</title>
        <script src="https://cdn.jsdelivr.net/npm/vega@5"></script>
        <script src="https://cdn.jsdelivr.net/npm/vega-lite@4"></script>
        <script src="https://cdn.jsdelivr.net/npm/vega-embed@6"></script>
    </head>
    <body>
        <h1>Report</h1>
        <p>TODO describe data</p>
        {% for markup in result_markups %}
        {{ markup }}
        {% endfor %}
    </body>
</html>
